//=============================================================================
// xBanSecurity.
//	Author: Pravin P. Prabhu
//	Revision: 1.0
//	Last Modified: 1/21/09
//	Abstract:
//		Provides a simple interface that allows administrators to kickban
//	other players. Admins may ban arbitrary ranges as well. Do not attempt
//	to ban *.*.*.*!
//=============================================================================
class xBanSecurity expands MutatorPlus
  config(xBanSecurity);

#exec obj load file="..\sharedSystem\ipPlugin.u"
  
// Classwide constants
const COMMAND_PREFIX = "xbansecurity";	// Prefix for using xBanSecurity commands through mutate
const AUTHORITY_TABLE_NAME = 'xBanSecurity';  // Entry in the table

const CLASS_XBASECOMMANDS = Class'xBaseCommands';	// Class to find mutate call in
const CLASS_IPPLUGIN = Class'IPPlugin';

const BAD_PLUGIN = "Warning; bad IPPlugin class variable. Failed to load IPPlugin!";

const ASC_0 = 0x30;  // numerical consts
const ASC_9 = 0x39;

// Data structures
struct s_IP
{
  var() config string firstsubnet,secondsubnet,thirdsubnet,fourthsubnet; // firstsubnet.secondsubnet.thirdsubnet.fourthsubnet:port
};

struct s_banInfo // An IPv4 ban entry, consisting of four subnets and internal data about how this ban was recorded
{
  var() config s_IP IP;
  var() config string banEntryDate; // Timestamp of when the ban was inserted
  var() config string bannedBy; // Name of the admin who banned this player
  var() config string banReply; // What to tell this asshole when they try to connect (rejection message)
};


// Configurable variables
var(option) config s_banInfo banTable[1024];  // Up to 256 bans may be entered
var(option) config string IPPluginClass;  // IPPlugin to use with this mutator
var Class<IPPlugin> IPPlugin; // IPPlugin to use


// bullshit fixes for sweeney's idiocy
var bool bLoaded;

// Methods
/* postBeginPlay
 * Override to load the correct IPPlugin
 */
simulated function postBeginPlay()
{
  if( bloaded ) 
    return;
  bloaded=true; // Load only ONCE.

  super.postBeginPlay();
  
  if( level.netmode != nm_client )
  {
    saveConfig(); // Generate a config file
  
    // Try to use the correct IPPlugin class.
    IPPlugin = Class<IPPlugin>(DynamicLoadObject(IPPluginClass,class'CLASS'));
  
    if( IPPlugin == none )
      Warn(BAD_PLUGIN);
  }
}

/* Mutate
 * Override to provide forwarding commands to the server
 */
simulated function Mutate( Pawn Sender, Coerce string data )
{
local xBaseCommands XBC;	// Reference to the player's current xBaseCommands actor

  If( !(Left( data, len(COMMAND_PREFIX) ) ~= COMMAND_PREFIX ) )
	return;
		
  Data = CAPS(Data);
  XBC = xBaseCommands(Sender.FindInventoryType(CLASS_XBASECOMMANDS));
  XBC.ServerMutate( Data );
}

/* ServerMutate: Server-side command processing
 * Parameters: Sender of the command, the command itself, and mutator to forward to (DEPRECATED, DO NOT USE)
 *
 * Abstract: 
 *		Analyzes the incoming line of data (the command) and parses it accordingly.
 */
Function ServerMutate( Pawn Sender, Coerce String Data, optional MutatorPlus ForwardTO )
{
local string err;	// Error code (dummy variable)
local int offset;  // Offset when finding the index of a character in the commandline
local string command; // Parsed command
local string options; // options to pass to command
	
  // Validity checks - see if user has authority to use this
  If( !(Left( Data, len(COMMAND_PREFIX) ) ~= COMMAND_PREFIX ) 
 	|| !xCoop(Level.Game).HasAuthority( Playerpawn(Sender), AUTHORITY_TABLE_NAME, Err ) )
  	return;
		
  // Reduce the data to be just the command to be executed. This will be analyzed soon.
  Data = Right( Data, len(Data) - Len(COMMAND_PREFIX) );
  Data = Caps(Data);		// Remove case sensitivity
  Trim(Data); // Remove extranneous spaces
	
  // See what command it is
  command = parseNextWord( Data, " " );
	
  switch( command )
  {
	case("BANID"):
	  banID( sender, data );
	  break;

    case("BANIP"):
      banIP( sender, data );
      break;

    case("LISTBANS"):
      listBans( sender );
      break;

    case("REMOVEBAN"):
      deleteBan( sender, data );
      break;
      
    case("DISPLAYIPS"):
      displayIPs( sender );
      break;

    case("XBANSECURITY"):
      sender.clientMessage("xBanSecurity:",'ServerMutate');
      sender.clientMessage("Mutate xBanSecurity DisplayIPs - Displays the IP of each player currently connected to the server");
      sender.clientMessage("Mutate xBanSecurity BanID <ID> <Rejection Message> - Ban a player with the ID specified; when they try to reconnect, they will be given the rejection message you specify.");
      sender.clientMessage("Mutate xBanSecurity BanIP <IP> - Ban a specific IP address or range. Be sure to format the IP as follows: #.#.#.#:# or #.#.#.#");
      sender.clientMessage("Mutate xBanSecurity ListBans - Prints a list of all bans currently in the ban table.");
      sender.clientMessage("Mutate xBanSecurity RemoveBan <Slot> - Remove a specific ban entry from the ban table.");
      break;
     
    default:
      sender.clientMessage("Improper syntax. Use Mutate xBanSecurity xBanSecurity for help.",'ServerMessage');
  }
}

/* banID
 * Bans a player currently in the server. Also kicks them out after the ban is complete.
 */
function bool banID( pawn administrator, string options )
{
local string firstChar; // The first character in the string
local pawn banned; // The player who is about to get banned
local string IP; // IP of the offending player
local int banID; // ID of the offender

  // See if we have the ability to even get IP info
  if( IPPlugin == None )
  {
    Warn(administrator@"tried to issue a ban, but there is no valid IPPlugin!");
    return false;
  }

  // Assume options is a numerical string
  firstChar = caps(Left(options,1));
  
  if( !isNumber(firstChar) )
  {
    // invalid. Not 0 - 9 inclusive
    administrator.clientmessage("BanID command failed. Use BanID(space)ID(space)Message",'ServerMessage');
    return false;
  }
  
  banID = int( parseNextWord( options ) );
  
  banned = xCoop(level.game).getPlayer( banID ); // Find the player associated with this ID
  if( banned == none || !banned.isa('playerpawn') )
  {
    administrator.clientmessage("BanID command failed. You can only ban players...",'ServerMessage');
    log(administrator@"attempted to ban ID"@int(options));
    return false;
  }
  
  // Get his info
  IP = IPPlugin.static.retrieveIP( level, playerpawn(banned) );
  
  // LET THE MIGHTY BANHAMMER FALL!!!
  if( insertBan( administrator, IP, parseNextWord(options,"\\") ) )
  {
    // Success    
    banned.destroy(); // Kick them out
    xCoop(level.game).adminCommand( playerpawn(administrator),"issued a ban against"@banned.getHumanName() );
  }
  else
  {
    // Failure
    administrator.clientmessage("BanID failed. Parsing error occurred when evaluating player's IP.",'ServerMessage');
  }
}

/* banIP
 * Bans a specific IP or range by directly attempting a ban insert through the insertBan method.
 */
function bool banIP( pawn administrator, string options )
{
  if( insertBan( administrator, options ) )
  {
    xCoop(level.game).adminCommand( playerpawn(administrator),"inserted an IP ban, IP and reason:"@options );
  }
  else
  {
    xCoop(level.game).adminCommand( playerpawn(administrator), "attempted to issue an IP ban, IP and reason:"@options );
    administrator.clientmessage("BanIP failed. Use BanIP(space)#.#.#.#  The wildcard * is also usable.",'ServerMessage');
  }
}

/* insertBan 
 * Inserts a new ban in IPv4 dotted IP format with optional rejection message
 */
function bool insertBan( pawn administrator, string IP, optional string newRejectionMessage )
{
local string firstSub,secondSub,thirdSub,fourthSub; // Subnets of the IP
local int index;  // For array traversal of the ban table

  // Validity check: see if this is a valid IPv4 IP address.
  firstSub = parseNextWord(IP,".");
  secondSub = parseNextWord(IP,".");
  thirdSub = parseNextWord(IP,".");
  fourthSub = parseNextWord(IP,":");
  if( inStr( fourthSub, " " ) > -1 )
    fourthSub = left( fourthSub, inStr( fourthSub, " " ) );
  
  if( firstSub == "" || (firstSub != "*" && !isNumber(left(firstSub,1) ) ) || 
      secondSub == "" || (secondSub != "*" && !isNumber(left(secondSub,1) ) ) || 
      thirdSub == "" || (thirdSub != "*" && !isNumber(left(thirdSub,1) ) ) || 
      fourthSub == "" || (fourthSub != "*" && !isNumber(left(fourthSub,1) ) ) || 
      parseNextWord(IP) != "" )
  {
    // invalid
    return false;
  }
  
  // perform actual insertion
  for( index = 0; index < arraycount(banTable); index++ )
  {
    if( banTable[index].banEntryDate == "" )
    {
      // This is empty, insert here
      banTable[index].IP.firstSubnet = firstSub;
      banTable[index].IP.secondSubnet = secondSub;
      banTable[index].IP.thirdSubnet = thirdSub;
      banTable[index].IP.fourthSubnet = fourthSub;
      
      if( playerpawn(administrator) != none )
      	banTable[index].bannedBy
      	 = administrator.getHumanName()@"("$IPPlugin.static.retrieveIP(level,playerpawn(administrator))$")";
      else
        banTable[index].bannedBy = "(automated)";
        
      banTable[index].banReply = newRejectionMessage;
      
      banTable[index].banEntryDate
       = Level.Month$"/"$Level.Day$"/"$Level.Year$" ("$Level.Hour$":"$Level.Minute$")";
       
      saveConfig();
      
      Log(">>       Alert       <<",'xBanSecurity');
      Log(">>  New ban issued  << - Admin "$administrator.getHumanName()@"banned"@firstSub$"."$secondSub$"."$thirdSub$"."$fourthSub@"at slot"@index,'xBanSecurity');
      Log(">>       Alert       <<",'xBanSecurity');
      return true; // Successful ban
    }
  }
  
  Log("Ban table exhausted! Clean ban table or increase array size!",'xBanSecurity');
  return false;
}

/* removeBan
 * Removes a ban entry from a slot
 */
function bool removeBan( pawn administrator, int slot )
{
  if( slot < 0 || slot >= arrayCount( banTable ) )
  {
    administrator.clientMessage("Cannot unban slot "$slot$". Make sure you are attempting to unban the correct slot.",'ServerMessage');
    return false; // invalid
  }
  
  banTable[slot].IP.firstSubnet = "";
  banTable[slot].IP.secondSubnet = "";
  banTable[slot].IP.thirdSubnet = "";
  banTable[slot].IP.fourthSubnet = "";
  
  banTable[slot].bannedBy = "";
  
  banTable[slot].banEntryDate = "";
  
  banTable[slot].banReply = "";
  
  saveConfig();
    
  Log(">>       Alert       <<",'xBanSecurity');
  Log(">> Ban entry removed << - "$administrator.getHumanName()@"unbanned slot"@slot,'xBanSecurity');
  Log(">>       Alert       <<",'xBanSecurity');
  return true; // Successful remove
}

/* listBans
 * Prints a list of all bans on administrator's console
 */
function listBans( pawn administrator )
{
local int index; // for array traversal

  for( index = 0; index < arraycount( bantable ); index++ )
  {
    if( banTable[index].banEntryDate != "" )
    {
      administrator.clientMessage("Ban #"$index@banTable[index].IP.firstSubnet$"."$banTable[index].IP.secondSubnet$"."$banTable[index].IP.thirdSubnet$"."$banTable[index].IP.fourthSubnet,'ServerEvent');
      administrator.clientMessage("Banned on "$banTable[index].banEntryDate$" by "$banTable[index].bannedBy@"with message"@banTable[index].banReply,'ServerEvent');
    }
  }
}

/* displayIPs
 * Prints a list of all currently connected players' IPs to this administrator
 */
function displayIPs( pawn administrator )
{
local pawn pending; // For iteration

  administrator.clientMessage("IP List:",'serverEvent');

  for( pending = level.pawnList; pending != none; pending = pending.nextPawn )
  {
    if( pending.isA('PlayerPawn') )
      administrator.clientMessage(pending.getHumanName()$" - "$IPPlugin.static.retrieveIP(level,playerpawn(pending)),'ServerEvent');
  }
}

/* deleteBan
 * Called by players. Request to remove a certain ban from the ban table.
 */
function deleteBan( pawn administrator, string options )
{
local int slot; // proposed slot to remove the ban from 

  if( !isNumber( left( options, 1 ) ) )
  {
    administrator.clientMessage("Invalid ban table slot. Use removeBan(space)<int slot>",'ServerEvent');
    return;
  } 
  
  slot = int(parseNextWord(options));
  
  if( removeBan( administrator, slot ) )// Perform removal
  {
    xCoop(level.game).adminCommand( playerpawn(administrator), "removed a ban from slot"@slot );
  }
  else
  {
    xCoop(level.game).adminCommand( playerpawn(administrator), "attempted to removed a ban from slot"@slot );
  }
}

/* MutatePreLogin
 * Override to check if an incoming player is banned.
 */
function bool MutatePreLogin( String Options, String Password, String PlayerName, out string Error )
{
  return !isBanned(none,Error);  // Get last prelogin IP
}

/* isBanned
 * Returns true if Other is banned, false otherwise 
 */
function bool isBanned( playerpawn Other, out string rejectionMessage )
{
  local int index;  // For array traversal  
  local string IPResult;  // Other's IP in string format'
  local string parse; // Temp result of parsing
  
  local s_IP parsedIP; // Four separated subnets in one structure.. parsed version of IPResult
  
  if( IPPlugin == none )
  {
    if( other != none )
      warn("Could not assess ban status of"@other$".. Running xBanSecurity without a valid IPPlugin is pointless!");
    else 
      warn("Running xBanSecurity without a valid IPPlugin is pointless!");
    return false;
  }
  
  IPResult = IPPlugin.Static.retrieveIP(Level,Other); // get ip
  
  parsedIP.firstSubnet = parseNextWord( IPResult, "." );
  parsedIP.secondSubnet = parseNextWord( IPResult, "." ); // Get second subnet
  parsedIP.thirdSubnet = parseNextWord( IPResult, "." ); // Get first subnet
  parsedIP.fourthSubnet= parseNextWord( IPResult, ":" );  // get fourth subnet

  for(index=0;index<arraycount(banTable);index++)
  {
    if( banTable[index].banEntryDate != "" )
    {
      // Check against this ban
      if( (banTable[index].IP.firstSubnet == "*" || int(banTable[index].IP.firstSubnet) == int(parsedIP.firstSubnet)) &&
      (banTable[index].IP.secondSubnet == "*" || int(banTable[index].IP.secondSubnet) == int(parsedIP.secondSubnet)) &&
      (banTable[index].IP.thirdSubnet == "*" || int(banTable[index].IP.thirdSubnet) == int(parsedIP.thirdSubnet)) &&
      (banTable[index].IP.fourthSubnet == "*" || int(banTable[index].IP.fourthSubnet) == int(parsedIP.fourthSubnet)) )
      {
        // This jackass is banned
        if( banTable[index].banReply == "" )
          rejectionMessage = "You are banned from this server. Contact the administrator for further details.";
        else
          rejectionMessage = banTable[index].banReply;
        
        // Log and broadcast this event
        log("Incoming connection from "$banTable[index].IP.firstSubnet$"."$banTable[index].IP.secondSubnet$"."$banTable[index].IP.thirdSubnet$"."$banTable[index].IP.fourthSubnet@"was rejected. Reason:"@rejectionMessage,'xBanSecurity');
        broadcastMessage("Incoming connection from "$banTable[index].IP.firstSubnet$"."$banTable[index].IP.secondSubnet$"."$banTable[index].IP.thirdSubnet$"."$banTable[index].IP.fourthSubnet@"was rejected. Reason:"@rejectionMessage,true,'ServerEvent');
          
        return true;
      }
    }
  }
}

/* parseNextWord
 * Parses off the next word in the stream, delimited by a certain character. Note that
 * the commandline IS altered - the word is removed entirely and returned.
 */
function string parseNextWord( out string commandline, optional string delimitor )
{
  local int index; // index of the space in the string
  local string retval; // Return value
  
  if( delimitor == "" )
    delimitor = " "; // Default to space
    
  trim( commandline );
  
  index = instr(commandline,delimitor);
  if( index == -1 )
  {
    retval = commandline;
    commandline = "";
    return retval;
  }

  retval = left( commandline, index );
  commandline = right( commandline, len(commandline) - index - 1 ); // Remove the word from the commandline
  return retval;
}

/* isNumber
 * Simply returns true if the character is numerical, false if not 
 */
function bool isNumber( string input )
{
  return (asc(input) >= ASC_0 && asc(input) <= ASC_9);
}

/* Helper bs functions 
	Mostly from UnrealWiki.. well, entirely from UnrealWiki. */
static Function LTrim( Out String S )
{
	While( Len(S) > 0 && Left(S,1) == " " )
		S = Right(S,Len(S) -1);
}

static Function RTrim( out string S )
{
	While( Len(S) > 0 && Right( S, 1 ) == " " )
		S = Left( S, Len(S) -1);
}

static Function Trim( Out String S )
{
	LTrim(S);
	RTrim(S);
}

defaultproperties
{
}
